/**CFile***********************************************************************

   FileName    [main.c]

   PackageName [poext]

   Synopsis    [Main program for the poext program.]

   Description []

   SeeAlso     []

   Author      [Congguang Yang]

  Copyright   [The file was created for academic use only in Department of 
  Electrical and Computer Engineering, University of Massachusetts, Amherst. 

  IN NO EVENT SHALL THE UNIVERSITY OF MASSACHUSETTS BE LIABLE FOR ANY DAMAGE
  CAUSED BY THE USE OF THIS SOFTWARE.]

******************************************************************************/

#include "bnet.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int BDDlopt_PrintNode ARGS((BnetNetwork*, BnetNode*, FILE*));
static char * BDDlopt_GetFileHead ARGS((char *));
static void printList ARGS((FILE*, char**, int));
static int  BDDlopt_PrintNodeRecursive ARGS((BnetNetwork*,BnetNode*,FILE*,st_table*));
static void loptPrintNode ARGS((BnetNode*, FILE*));
static int loptFindPiRecursive ARGS((BnetNetwork*,BnetNode*,st_table*));

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Main program for poext]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
int
main(
  int  argc,
  char ** argv)
{
    FILE	*fp;		/* network file pointer */
    BnetNetwork	*net = NULL;	/* network */
    int		exitval;	/* return value of Cudd_CheckZeroRef */
    int		ok;		/* overall return value from main() */
    int		result;		/* stores the return value of functions */
    BnetNode	*node;		/* auxiliary pointer to network node */
    BnetNode    *POnode;	/* New variable, to retrieve BDD manager from this PO node */
    int		i;		/* loop index */
    int		j;		/* loop index */
    int		k;
    int		level;		/* aux. var. used to print variable orders */
    char	*filename;	/* Input blif file_name */
    char	*dfilename;	/* Dump file names */
    FILE	*dfp;
    char	*dump_name;

    if (argc < 2) {
 	printf("Usage: poext ckt.blif\n");
	exit(2);
    }

    /* Initialize. */
    filename = util_strsav(argv[1]);

    /* Read the Boolean network */
    fp = fopen(filename, "r");
    if (fp == NULL) {
        fprintf(stderr, "Circuit file read error\n");
	exit(2);
    }
    net = Bnet_ReadNetwork(fp,1);
    fclose(fp);

    /* Print out PO one by one */
    for (i = 0; i < net->npos; i++) {
	if (!st_lookup(net->hash,net->outputs[i],(char **) &POnode)) {
	    exit(2);
	}

        /* Get rid of '(' or '<' in the PO name */
	dump_name = ALLOC(char, 20);
        k = 0;
	j = 0;
	while (POnode->name[j] != NULL) {
	     if (POnode->name[j] != '(' && POnode->name[j] != ')' 
		 		        && POnode->name[j] != '<' && POnode->name[j] != '>') {
	         dump_name[k] = POnode->name[j];
		 k++;
 	     }
	     j++;
	}
	dump_name[k] = '\0';
	
    	/* Built dump file name */
	dfilename = ALLOC(char, 100);
	sprintf(dfilename,"%s_%s.blif",BDDlopt_GetFileHead(filename),dump_name);
  	dfp = fopen(dfilename, "w");

    	result = BDDlopt_PrintNode(net, POnode, dfp);
    	if (result == 0) { exit(2); }

   	fclose(dfp);
	FREE(dfilename);
	FREE(dump_name);
    } 

    Bnet_FreeNetwork(net);

    exit(ok);

} /* end of main */

/**Function********************************************************************

  Synopsis    [Get the file head]

  Description []

  SideEffects [None]

  SeeAlso     []

*****************************************************************************/
static
char *
BDDlopt_GetFileHead(
  char *fname)
{
  int i = 0;
  char *file_head, *q;
  FILE *tstf;

  /* Test if the file exists or not */
  if ((tstf = fopen(fname,"r")) == NULL) {
      perror(fname);
      exit(2);
  } else {
      fclose(tstf);
  }

  file_head = ALLOC(char, 20); /* The maximum length of the name of a test case */
  for (q = fname; *q != '.'; q++) {
      file_head[i] = *q;
      i++;
  }
  file_head[i] = '\0';

  return(file_head);

} /* end of BDDlopt_GetFileHead */

/**Function********************************************************************

  Synopsis    [Dump a PO node in blif format]

  Description []

  SideEffects [None]

  SeeAlso     []

*****************************************************************************/
static
int
BDDlopt_PrintNode(
  BnetNetwork *net,
  BnetNode *nd,
  FILE *dfp)
{
  int i, result;
  st_table *visited;	/* Node visited */
  st_table *Pi; 	/* PI used in this PO */
  BnetNode *input;

  visited = st_init_table(st_ptrcmp,st_ptrhash);
  Pi = st_init_table(st_ptrcmp,st_ptrhash);

  fprintf(dfp,".model %s_%s\n", net->name, nd->name);

  /* Traverse network to find out Pi used in this PO */
  for (i = 0; i < nd->ninp; i++) {
      if (!st_lookup(net->hash,(char *) nd->inputs[i], (char **) &input)) {
	  printf("Can not find input nodes !");
	  return(0);
      }
      result = loptFindPiRecursive(net, input, Pi);
      if (result == 0) return(0);
  }

  /* Print out Pi used 
  */
  fprintf(dfp,".inputs");
  for (i = 0; i < net->npis; i++) {
      if (!st_lookup(net->hash,net->inputs[i],(char **) &input)) {
	  return(0);
      }

      if (st_is_member(Pi, (char *) input) == 1) { 
	  fprintf(dfp," %s", input->name);
      }
  }
  fprintf(dfp, "\n");

  fprintf(dfp,".outputs %s\n", nd->name);


  if (nd->type == BNET_CONSTANT_NODE || nd->type == BNET_INPUT_NODE) {
      return(1);
  }
  else { /* nd->type == BNET_INTERNAL_NODE or BNET_OUTPUT_NODE */

      /* Print out this node */
      loptPrintNode(nd, dfp);

      if (st_add_direct(visited, (char *)nd, NULL) == ST_OUT_OF_MEM) {
	  printf("Out of memory !");
	  return(0);
      }

      /* Print inputs recursively */
      for (i = 0; i < nd->ninp; i++) {
	  if (!st_lookup(net->hash,(char *) nd->inputs[i], (char **) &input)) {
	      printf("Can not find input nodes !");
	      return(0);
	  }

          result = BDDlopt_PrintNodeRecursive(net, input, dfp, visited);
          if (result == 0) return(0);
      }
  }

  fprintf(dfp,".end\n");

  st_free_table(Pi);
  st_free_table(visited);

  return(1);

} /* end of BDDlopt_PrintNode */

/**Function********************************************************************

  Synopsis    [Prints a list of strings to the standard output.]

  Description []

  SideEffects [None]

  SeeAlso     [readList Bnet_PrintNetwork]

******************************************************************************/
static 
void
printList(
  FILE *dfp,
  char ** list,
  int  n)
{
    int i;

    for (i = 0; i < n; i++) {
        fprintf(dfp," %s",list[i]);
    }
    fprintf(dfp,"\n");

} /* end of printList */

/**Function********************************************************************

  Synopsis    [Traverse network, print out nodes]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
static
int
loptFindPiRecursive(
  BnetNetwork *net, 
  BnetNode *nd, 
  st_table *Pi)
{
  int i, result;
  BnetNode *input;

  if (nd->type == BNET_CONSTANT_NODE) {
      return(1);
  }
  else if (nd->type == BNET_INPUT_NODE && st_is_member(Pi, (char *) nd) != 1) {

      if (st_add_direct(Pi, (char *)nd, NULL) == ST_OUT_OF_MEM) {
	  printf("Out of memory !");
	  return(0);
      }
      return(1);
  }
  else { /* INTERNAL NODES */
    
      for (i = 0; i < nd->ninp; i++) {
	  if (!st_lookup(net->hash,(char *) nd->inputs[i], (char **) &input)) {
	      printf("Can not find input nodes !");
	      return(0);
          }
      
          result = loptFindPiRecursive(net, input, Pi);
          if (result == 0) return(0);
      }
  }

  return(1);

} /* end of loptFindPiRecursive */

/**Function********************************************************************

  Synopsis    [Traverse network, print out nodes]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
static 
int  
BDDlopt_PrintNodeRecursive(
  BnetNetwork *net,
  BnetNode *nd,
  FILE *dfp,
  st_table *visited)
{
  int i, result;
  BnetNode *input;

  if (nd->type == BNET_CONSTANT_NODE || nd->type == BNET_INPUT_NODE) {
      return(1);
  }
  else if (st_is_member(visited, (char *) nd) == 1) { 
      return(1);
  }
  else { /* Never printed before */

      /* Print out this node */
      loptPrintNode(nd, dfp);

      if (st_add_direct(visited, (char *)nd, NULL) == ST_OUT_OF_MEM) {
	  printf("Out of memory !");
	  return(0);
      }

      /* Print inputs recursively */
      for (i = 0; i < nd->ninp; i++) {
	  if (!st_lookup(net->hash,(char *) nd->inputs[i], (char **) &input)) {
	      printf("Can not find input nodes !");
	      return(0);
          }
      
          result = BDDlopt_PrintNodeRecursive(net, input, dfp, visited);
          if (result == 0) return(0);
      }
  }

  return(1);

} /* end of BDDlopt_PrintNodeRecursive */

/**Function********************************************************************

  Synopsis    [Prints a node]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
static
void
loptPrintNode(
  BnetNode *nd,
  FILE *dfp)
{  
  int i;
  BnetTabline *tl;

  fprintf(dfp,".names");
  for (i = 0; i < nd->ninp; i++) {
      fprintf(dfp," %s",nd->inputs[i]);
  }

  fprintf(dfp," %s\n",nd->name);

  tl = nd->f;
  while (tl != NULL) {
      if (tl->values != NULL) 
          fprintf(dfp,"%s %d\n",tl->values, 1 - nd->polarity);
      else 
          fprintf(dfp,"%d\n", 1 - nd->polarity);
     
      tl = tl->next;
  }

} /* end of loptPrintNode */

